def tick_from_bitmap(index, num, tick_spacing: int) -> list[int]:
    bin_str = bin(num)[2:][::-1]
    initialized = []
    for idx, val in enumerate(bin_str):
        if val == "1":
            initialized.append(idx)
    return [(index * 256 + i) * tick_spacing for i in initialized]



from .constants import UniswapV2FactoryAddresses
from .events import (
    V2BurnEvent,
    V2BurnEventType,
    V2MintEvent,
    V2MintEventType,
    V2PairCreatedEvent,
    V2PairCreatedEventType,
    V2SwapEvent,
    V2SwapEventType,
    V2SyncEvent,
    V2SyncEventType,
)
from .factory import GetPairRequest, UniswapV2Factory
from .pair import UniswapV2Pair
from .router import UniswapV2Router

__all__ = [
    "GetPairRequest",
    "UniswapV2Factory",
    "UniswapV2FactoryAddresses",
    "UniswapV2Pair",
    "UniswapV2Router",
    "V2SwapEventType",
    "V2SyncEventType",
    "V2PairCreatedEventType",
    "V2SwapEvent",
    "V2PairCreatedEvent",
    "V2SyncEvent",
    "V2MintEvent",
    "V2MintEventType",
    "V2BurnEvent",
    "V2BurnEventType",
]



from typing import Any

from eth_rpc.types import Bytes32Hex, EIP712Model, primitives
from pydantic import ConfigDict, Field


def get_tweet_timestamp(tid: int) -> int:
    offset = 1288834974657
    tstamp = (tid >> 22) + offset
    return int(tstamp)


class UserMention(EIP712Model):
    id: Bytes32Hex
    screen_name: str = Field(alias="screenName")
    start_index: primitives.uint256 = Field(alias="startIndex")
    end_index: primitives.uint256 = Field(alias="endIndex")

    model_config = ConfigDict(populate_by_name=True)

    @staticmethod
    def from_mention(mention: dict[str, Any]):
        return UserMention(
            id=mention["id_str"],
            screen_name=mention["screen_name"],
            start_index=mention["indices"][0],
            end_index=mention["indices"][1],
        )


class Tweet(EIP712Model):
    id: Bytes32Hex
    user_id: Bytes32Hex = Field(alias="userId")
    created_at: primitives.uint64 = Field(alias="createdAt")
    conversation_id: Bytes32Hex = Field(alias="conversationId")
    full_text: primitives.string = Field(alias="fullText")
    in_reply_to_status_id: Bytes32Hex = Field(
        alias="inReplyToStatusId", default=primitives.bytes32(b"")
    )
    in_reply_to_user_id: Bytes32Hex = Field(
        alias="inReplyToUserId", default=primitives.bytes32(b"")
    )
    is_quote: bool = Field(alias="isQuote")
    mentions: list[UserMention] = Field(default_factory=list)

    model_config = ConfigDict(populate_by_name=True)

    @property
    def id_int(self):
        return int(self.id.hex(), 16)



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives
from pydantic import BaseModel, Field


class DeployArgs(BaseModel):
    owner_id: primitives.bytes32
    namespace: primitives.bytes32 = Field(default=primitives.bytes32(b""))
    index: primitives.uint256 = Field(default=primitives.uint256(0))


class AccountManager(ProtocolBase):
    deploy: ContractFunc[
        tuple[DeployArgs],
        primitives.address,
    ] = METHOD

    compute_address: Annotated[
        ContractFunc[
            DeployArgs,
            primitives.address,
        ],
        Name("computeAddress"),
    ] = METHOD



from abc import ABC, abstractmethod
from typing import Annotated, Any, ClassVar

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from pydantic import BaseModel


class SimulacrumSubmission(BaseModel):
    source_data: bytes
    verification_data: bytes = b""
    require_success: bool = True


class BaseSource(ProtocolBase, ABC):
    name: ClassVar[str]
    input_type: ClassVar[type]

    submit: ContractFunc[
        SimulacrumSubmission,
        Annotated[primitives.bool, Name("success")],
    ] = METHOD

    submit_many: Annotated[
        ContractFunc[
            tuple[list[bytes], list[bytes], bool],
            None,
        ],
        Name("submitMany"),
    ] = METHOD

    registry: ContractFunc[
        NoArgs,
        primitives.address,
    ] = METHOD

    @classmethod
    @abstractmethod
    def validate_input(cls, input: Any): ...



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import Name, primitives
from pydantic import BaseModel


class Route(BaseModel):
    from_: Annotated[primitives.address, Name("from")]
    to: primitives.address
    stable: bool


class EthSwapRequest(BaseModel):
    amount_out_min: primitives.uint256
    route: list[Route]
    to: primitives.address
    deadline: primitives.uint256


class TokenSwapRequest(BaseModel):
    amount_in: primitives.uint256
    amount_out_min: primitives.uint256
    route: list[Route]
    to: primitives.address
    deadline: primitives.uint256


class LynexV2Router(ProtocolBase):
    swap_exact_tokens_for_tokens_supporting_fee_on_transfer_tokens: Annotated[
        ContractFunc[
            TokenSwapRequest,
            None,
        ],
        Name("swapExactTokensForTokensSupportingFeeOnTransferTokens"),
    ]

    swap_exact_eth_for_tokens_supporting_fee_on_transfer_tokens: Annotated[
        EthSwapRequest,
        Name("swapExactETHForTokensSupportingFeeOnTransferTokens"),
    ]

    # TODO: finish adding methods



from .account_manager import AccountManager
from .blocks import Blocks, BlockStorage, NewBlockEvent, NonceArgs
from .helpers import TokenRegistry, TokenStruct
from .mempool import Mempool
from .registry import ProtocolRegistry
from .sources import CommandSource, XSource
from .utils import Command, CommandStatus

__all__ = [
    "AccountManager",
    "Blocks",
    "BlockStorage",
    "Command",
    "CommandStatus",
    "CommandSource",
    "Mempool",
    "NewBlockEvent",
    "NonceArgs",
    "ProtocolRegistry",
    "TokenRegistry",
    "TokenStruct",
    "XSource",
]



from eth_rpc import Event

from .types import ApprovalEventType, TransferEventType

TransferEvent = Event[TransferEventType](name="Transfer")
ApprovalEvent = Event[ApprovalEventType](name="Approval")



from typing import Annotated

from eth_rpc.event import Event
from eth_rpc.types import Indexed, primitives
from pydantic import BaseModel


class V3PoolCreatedEventType(BaseModel):
    token0: Annotated[primitives.address, Indexed]
    token1: Annotated[primitives.address, Indexed]
    fee: Annotated[primitives.uint24, Indexed]
    tick_spacing: primitives.int24
    pool: primitives.address


class V3SwapEventType(BaseModel):
    sender: Annotated[primitives.address, Indexed]
    recipient: Annotated[primitives.address, Indexed]
    amount0: primitives.int256
    amount1: primitives.int256
    sqrt_price_x96: primitives.uint160
    liquidity: primitives.uint128
    tick: primitives.int24


class V3MintEventType(BaseModel):
    sender: primitives.address
    owner: Annotated[primitives.address, Indexed]
    tick_lower: Annotated[primitives.int24, Indexed]
    tick_upper: Annotated[primitives.int24, Indexed]
    amount: primitives.uint128
    amount0: primitives.uint256
    amount1: primitives.uint256


class V3IncreaseLiquidityEventType(BaseModel):
    """From the NFT Position Manager"""

    token_id: Annotated[primitives.uint256, Indexed]
    liquidity: primitives.uint128
    amount0: primitives.uint256
    amount1: primitives.uint256


# FACTORY
V3PoolCreatedEvent = Event[V3PoolCreatedEventType](name="PoolCreated")

# POOL
V3SwapEvent = Event[V3SwapEventType](name="Swap")
V3MintEvent = Event[V3MintEventType](name="Mint")

# NFT_POSITION_MANAGER
V3IncreaseLiquidityEvent = Event[V3IncreaseLiquidityEventType](name="IncreaseLiquidity")



import asyncio
from typing import Annotated

from eth_rpc.types import Name, primitives
from eth_typeshed.constants import Factories
from eth_typeshed.erc20 import ERC20
from eth_typing import HexAddress
from pydantic import BaseModel, PrivateAttr

from .factory import GetPoolRequest, UniswapV3Factory
from .pool import UniswapV3Pool
from .utils import get_fees, liquidity_to_token_amounts


class OwnerTokenRequest(BaseModel):
    owner: HexAddress
    index: primitives.uint256


class Position(BaseModel):
    _pool_address: HexAddress = PrivateAttr(None)

    nonce: primitives.uint96
    operator: primitives.address
    token0: primitives.address
    token1: primitives.address
    fee: primitives.uint24
    tick_lower: Annotated[
        primitives.int24,
        Name("tickLower"),
    ]
    tick_upper: Annotated[
        primitives.int24,
        Name("tickUpper"),
    ]
    liquidity: primitives.uint128
    fee_growth_inside0_last: Annotated[
        primitives.uint256,
        Name("feeGrowthInside0LastX128"),
    ]
    fee_growth_inside1_last: Annotated[
        primitives.uint256,
        Name("feeGrowthInside1LastX128"),
    ]
    tokens_owed0: Annotated[
        primitives.uint128,
        Name("tokensOwed0"),
    ]
    tokens_owed1: Annotated[
        primitives.uint128,
        Name("tokensOwed1"),
    ]

    def pool(self, factory_address=Factories.Ethereum.UniswapV3):
        if not self._pool_address:
            self._pool_address = (
                UniswapV3Factory(address=factory_address)
                .get_pool(
                    GetPoolRequest(
                        token_a=self.token0,
                        token_b=self.token1,
                        fee=self.fee,
                    )
                )
                .get()
            )
        return UniswapV3Pool(address=self._pool_address)

    async def token_amounts(self):
        sqrt_price_x96 = (await self.pool().slot0().get()).sqrt_price
        return liquidity_to_token_amounts(
            self.liquidity,
            sqrt_price_x96,
            self.tick_lower,
            self.tick_upper,
        )

    async def get_tick_lower(self):
        return await self.pool().ticks(self.tick_lower).get()

    async def get_tick_upper(self):
        return await self.pool().ticks(self.tick_upper).get()

    async def price_range(self):
        token0_decimals = await ERC20(self.token0).decimals().get()
        token1_decimals = await ERC20(self.token1).decimals().get()
        lower = 1.0001**self.tick_lower * 10 ** (token0_decimals - token1_decimals)
        upper = 1.0001**self.tick_upper * 10 ** (token0_decimals - token1_decimals)
        return [lower, upper]

    async def get_pending_fees(self):
        pool = self.pool()
        (
            fee_growth_global0,
            fee_growth_global1,
            slot0,
            lower_tick,
            upper_tick,
        ) = await asyncio.gather(
            pool.fee_growth_global0().get(),
            pool.fee_growth_global1().get(),
            pool.slot0().get(),
            self.get_tick_lower(),
            self.get_tick_upper(),
        )

        return get_fees(
            fee_growth_global0,
            fee_growth_global1,
            lower_tick.fee_growth_outside0,
            upper_tick.fee_growth_outside0,
            self.fee_growth_inside0_last,
            lower_tick.fee_growth_outside1,
            upper_tick.fee_growth_outside1,
            self.fee_growth_inside1_last,
            self.liquidity,
            self.tick_lower,
            self.tick_upper,
            slot0.tick,
        )

    @property
    def range(self) -> tuple[int, int]:
        return self.tick_lower, self.tick_upper
